/*-----------------------------------------------------------------------

File  : ccl_rewrite.c

Author: Stephan Schulz

Contents
 
  Rewriting. For all functions, if index != NULL, delete rewritten
  terms and superterms from the index!

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Tue May 26 19:47:52 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_rewrite.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

long RewriteAttempts = 0;
long RewriteSucesses = 0;
/* long RewriteTermDepth = 0; */

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------
//
// Function: instance_is_rule()
//
//   Return true if lside->rside is a rule, i.e. lside>rside (for the
//   instantiated terms) and rside contains no unbound
//   variables. Assumes that uninstantated terms lside and rside are
//   uncomparable! 
//
// Global Variables: -
//
// Side Effects    : TOGreater() can manipulate TPOpFlag of terms.
//
/----------------------------------------------------------------------*/

static bool instance_is_rule(OCB_p ocb, Term_p lside, Term_p rside,
			     Subst_p subst)

{   
   if(TermHasUnboundVariables(rside))
   {
      return false;
   }
   if(SubstIsRenaming(subst)) /* Save comparisons */
   {
      return false;
   }
   return TOGreater(ocb, lside, rside, DEREF_ONCE, DEREF_ONCE);
}



/*-----------------------------------------------------------------------
//
// Function: term_is_top_rewritable()
//
//   Return true if the term is rewritable with new_demod at the top
//   position, false otherwise.
//
// Global Variables: -
//
// Side Effects    : Temporary variable bindings (backtracked)
//
/----------------------------------------------------------------------*/

static bool term_is_top_rewritable(OCB_p ocb, Term_p term, Clause_p
				   new_demod)
{
   Subst_p subst = SubstAlloc();
   Eqn_p   eqn;
   bool    res = false;
   
   assert(new_demod->pos_lit_no == 1);
   assert(new_demod->neg_lit_no == 0);
   assert(!TermIsVar(term));
   
   eqn = new_demod->literals;

   DEBUGMARK(RW_MATCH_WATCH, "Matching ");
   DEBUG(RW_MATCH_WATCH, 
	 ClausePrint(stdout, new_demod, true););
   DEBUGOUT(RW_MATCH_WATCH, " (l) onto ");
   DEBUG(RW_MATCH_WATCH, 
	 TermPrint(stdout, term, ocb->sig, DEREF_NEVER););
   DEBUGOUT(RW_MATCH_WATCH, "\nInstantiated lside: ");
   DEBUG(RW_MATCH_WATCH,
	 TermPrint(stdout, eqn->lterm, ocb->sig, DEREF_ONCE););
   DEBUGOUT(RW_MATCH_WATCH, "\n");
   if(SubstComputeMatch(eqn->lterm, term, subst, TBTermEqual))
   {
      if((EqnIsOriented(eqn) 
	  || instance_is_rule(ocb, eqn->lterm, eqn->rterm, subst))
	 &&
	 (!TermCellQueryProp(term, TPRestricted) ||
	  !SubstIsRenaming(subst)))
      {
	 res = true;
      }
      SubstBacktrack(subst);
   }
   if(!res && !EqnIsOriented(eqn))
   {
      if(SubstComputeMatch(eqn->rterm, term, subst, TBTermEqual))
      {
	 if(instance_is_rule(ocb, eqn->rterm, eqn->lterm, subst))
	    /* If instance is rule -> subst is no renaming! */
	 {
	    res = true;
	 }
      }
   }
   SubstDelete(subst);
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: term_is_rewritable()
//
//   Return true if the term is rewritable with new_demod, false
//   otherwise. Set nf_date[0,1] on non-rewritable terms to nf_date
//   (i.e. assumes thate term is in normal for with respect to earlier
//   systems).
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static bool term_is_rewritable(OCB_p ocb, Term_p term, Clause_p
			       new_demod, SysDate nf_date)
{
   int i;
   bool res = false;

   /* printf("term_is_rewritable()...\n"); */
   
   if(TermIsVar(term))
   {
      return false;
   }
   if(TermCellQueryProp(term, TPIsRewritable))
   {
      return true;
   }
   if(SysDateCompare(term->nf_date[RewriteAdr(FullRewrite)], nf_date) == DateEqual)
   {
      return false;
   }
   for(i=0; i<term->arity; i++)
   {
      if(term_is_rewritable(ocb, term->args[i], new_demod, nf_date))
      {
	 res = true;
	 break;
      }
   }
   if(res)
   {
      TermCellSetProp(term, TPIsRewritable);
      return true;
   }
   if(term_is_top_rewritable(ocb, term, new_demod))
   {
      TermCellSetProp(term, TPIsRewritable);
      return true;
   }
   term->nf_date[RewriteAdr(RuleRewrite)] = term->nf_date[RewriteAdr(FullRewrite)] = nf_date;
   
   /* printf("...term_is_rewritable() = false (no match)\n"); */
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: eqn_has_rw_side()
//
//   Return NoSide, MaxSide, MinSide depending on wether
//   eqn has or has not a rewritable (maximal) side.
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static EqnSide eqn_has_rw_side(OCB_p ocb, Eqn_p eqn, Clause_p
				new_demod, SysDate nf_date)
{
   bool resl, resr;

   resl = term_is_rewritable(ocb, eqn->lterm, new_demod, nf_date);
   resr = term_is_rewritable(ocb, eqn->rterm, new_demod, nf_date);

   if(resl)
   {
      return MaxSide;
   }
   else if(resr)
   {
      return EqnIsOriented(eqn)?MinSide:MaxSide;
   }
   return NoSide;
}
   
/*-----------------------------------------------------------------------
//
// Function: clause_has_rw_max_side()
//
//   Return MaxSide, if the clause has a rewritable  maximal literal that
//   may become non-maximal, MinSide if this is not the case, but it
//   has another rewritable, NoSide otherwise.
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static EqnSide clause_has_rw_max_side(OCB_p ocb, Clause_p clause,
				   Clause_p new_demod, SysDate
				   nf_date)
{
   Eqn_p handle;
   EqnSide tmp;
   bool maxres = false, minres = false;

   /* printf("Checking clause %ld: ", clause->ident);
      ClausePrint(stdout, clause, false);
   printf("\n"); */
   for(handle = clause->literals; handle; handle = handle->next)
   {
      tmp = eqn_has_rw_side(ocb, handle, new_demod, nf_date);
      if(tmp != NoSide)
      {
	 if(EqnIsMaximal(handle))
	 {	
	    if((tmp == MaxSide)||
	       (EqnIsPositive(handle) && (clause->pos_lit_no >=2))||
	       (EqnIsNegative(handle) && (clause->neg_lit_no >=2)))
	    {	  
	       maxres = true;
	    }
	    else
	    {
	       minres = true;
	    }
	 }
	 else
	 {
	    minres = true;
	 }
      }	 
   }
   if(maxres)
   {
      return MaxSide;
   }
   if(minres)
   {
      return MinSide;
   }
   return NoSide;
}
  
   
/*-----------------------------------------------------------------------
//
// Function: find_clauses_with_rw_max_sides()
//
//   A non-index-using implementation of
//   FindClausesWithRewritableMaxSides(). Return true if any clause
//   has a rewritable non-maximal side.
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static bool find_clauses_with_rw_max_sides(OCB_p ocb, ClauseSet_p set,
					   PTree_p* results, Clause_p
					   new_demod, SysDate nf_date)
{
   Clause_p handle;
   EqnSide  tmp;
   bool     res = false;
   
   assert(new_demod->pos_lit_no == 1);
   assert(new_demod->neg_lit_no == 0);
   
   for(handle = set->anchor->succ; handle != set->anchor; handle =
	  handle->succ)
   {
      tmp = clause_has_rw_max_side(ocb, handle, new_demod, nf_date);
      assert(tmp!=BothSides);
      if(tmp==MaxSide)
      {
	 PTreeStore(results, handle);
      }
      res = res || (tmp==MinSide);
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: indexed_find_demodulator()
//
//   Find a demodulator via demodulators->demod_index.
//
// Global Variables: -
//
// Side Effects    : Instantiates
//
/----------------------------------------------------------------------*/

static ClausePos_p indexed_find_demodulator(OCB_p ocb, Term_p term,
					    SysDate date,
					    ClauseSet_p demodulators,
					    Subst_p subst)
{
   Eqn_p       eqn;   
   ClausePos_p pos, res = NULL;
   
   DEBUGMARK(RW_INTERFACE_WATCH2,
	     "TermIndexedFindDemodulator()...\n");
   
   assert(term);
   assert(demodulators);
   assert(demodulators->demod_index);
   assert(term->weight == TermWeight(term, DEFAULT_VWEIGHT,
				     DEFAULT_FWEIGHT));

   RewriteAttempts++;   
   
   DEBUGMARK(RW_MATCH_WATCH, "\nSearching match for: ");
   DEBUG(RW_MATCH_WATCH, TermPrint(stdout, term, ocb->sig,
				   DEREF_NEVER);
	 printf("\n"););
   
   PDTreeSearchInit(demodulators->demod_index, term, date);
   
   while((pos = PDTreeFindNextDemodulator(demodulators->demod_index, subst)))
   {
      eqn = pos->literal;
      
      if((EqnIsOriented(eqn)&&
	  SysDateCompare(term->nf_date[RewriteAdr(RuleRewrite)],
			 pos->clause->date)!=DateEarlier)
	 ||
	 (!EqnIsOriented(eqn)&&
	  SysDateCompare(term->nf_date[RewriteAdr(FullRewrite)],
			 pos->clause->date)!=DateEarlier))
      {
	 continue;
      }

      switch(pos->side)
      {
      case LeftSide:
	    DEBUGMARK(RW_MATCH_WATCH, "Found: ");
	    DEBUG(RW_MATCH_WATCH, 
		  ClausePrint(stdout, pos->clause, true);
		  printf(" (l), instantiated: ");
		  TermPrint(stdout, pos->clause->literals->lterm,
			    ocb->sig, DEREF_ONCE);
		  printf(" <=> ");
		  TermPrint(stdout, pos->clause->literals->rterm,
			    ocb->sig, DEREF_ONCE);
		  printf("\n");
	       );
	    if((EqnIsOriented(eqn) 
		|| instance_is_rule(ocb, eqn->lterm, eqn->rterm, subst))
	       &&
	       (!TermCellQueryProp(term, TPRestricted) ||
		!SubstIsRenaming(subst)))
	    {
	       res = pos;
	    }
	    break;
      case RightSide:
	    DEBUGMARK(RW_MATCH_WATCH, "Found: ");
	    DEBUG(RW_MATCH_WATCH, 
		  ClausePrint(stdout, pos->clause, true);
		  printf(" (r), instantiated: ");
		  TermPrint(stdout, pos->clause->literals->lterm,
			    ocb->sig, DEREF_ONCE);
		  printf(" <=> ");
		  TermPrint(stdout, pos->clause->literals->rterm,
			    ocb->sig, DEREF_ONCE);
		  printf("\n");
	       );
	    assert(!EqnIsOriented(eqn));
	    if(instance_is_rule(ocb, eqn->rterm, eqn->lterm, subst)
	       &&
	       !TermCellQueryProp(term, TPRestricted)) 
	       /* Case SubstIsRenaming(subst) already eleminated in
		  instance_is_rule! */
	    {
	       res = pos;
	    }
	    break;
      default: 
	    assert(false);
	    break;
      }
      if(res)
      {
	 break;
      }
   }
   PDTreeSearchExit(demodulators->demod_index);

   DEBUGMARK(RW_INTERFACE_WATCH2,
	     "...TermIndexedFindDemodulator()\n");
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: rewrite_with_clauseset()
//
//   Rewrite the given term at root position with the first matching,
//   orientable rule from demodulators. Return true if a matching rule
//   was found, false otherwise.
//
// Global Variables: -
//
// Side Effects    : Changes term and termbank, by TBTermReplace()
//
/----------------------------------------------------------------------*/

static bool rewrite_with_clause_set(OCB_p ocb, TB_p bank, Term_p term,
				    SysDate date, ClauseSet_p demodulators)
{
   bool        res   = false;
   Subst_p     subst = SubstAlloc();
   ClausePos_p pos;
   Term_p      repl;

   assert(demodulators->demod_index);
   DEBUGMARK(RW_INTERFACE_WATCH2, "TermRewriteWithClauseSet()...\n");   

   pos = indexed_find_demodulator(ocb, term, date, demodulators, subst);
   if(pos)
   {
      DEBUG(RW_REWRITE_WATCH, printf("Rewriting %ld ", term->entry_no););
      /* DEBUG(RW_REWRITE_WATCH,
	    TermPrint(stdout, term, bank->sig, DEREF_NEVER);); */
      DEBUGOUT(RW_REWRITE_WATCH, " with ");
      DEBUG(RW_REWRITE_WATCH,
	    ClausePrint(stdout, pos->clause, true);
	    printf("(%c)\n", pos->side == LeftSide? 'l':'r'););
      
      RewriteSucesses++;

      repl = ClausePosGetOtherSide(pos);      
      repl = TBInsertInstantiated(bank, repl);
      DEBUGOUT(RW_REWRITE_WATCH, "Inserted instantiated right hand side\n");
      if(OutputLevel>=3)
      {
	 fprintf(GlobalOut, RW_STANDARD " %ld", pos->clause->ident);
      }
      TBTermReplace(bank, term, repl);
      DEBUGOUT(RW_REWRITE_WATCH, "Replacing ok\n");
      if(OutputLevel>=3)
      {
	 fputs(".\n", GlobalOut);
      }
      
      DEBUGMARK(RW_REWRITE_WATCH, "Result ");
      /* DEBUG(RW_REWRITE_WATCH,
	    TermPrint(stdout, repl, bank->sig, DEREF_NEVER););*/
      DEBUGOUT(RW_REWRITE_WATCH, "\n");
      
      res = true;
   }
   SubstDelete(subst);   

   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermRewriteWithClauseSet()\n");   
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: rewrite_with_clause_setlist()
//
//   Rewrite a term with the sets of demodulators.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static bool rewrite_with_clause_setlist(OCB_p ocb, TB_p bank, Term_p term,
					ClauseSet_p* demodulators,
					RewriteLevel level)
{
   unsigned int  i;
   bool res = false;

   assert(level);

   for(i=0; i<level; i++)
   {
      assert(demodulators[i]);
      if(SysDateCompare(term->nf_date[level-1], demodulators[i]->date) ==
	 DateEarlier)
      {
	 res = rewrite_with_clause_set(ocb, bank, term,
				       term->nf_date[level-1],
				       demodulators[i]);
	 if(res)
	 {
	    break;
	 }
      }
   }
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: term_LI_normalform()
//
//   Compute a leftmost-innermost normal form of term. This is more
//   tricky than expected, as supertems on the stack may change. I'll
//   start with a recursive version and see how it works. This uses
//   dates to minimize rewrite-attempts: If the normal form of the
//   term is younger than the clause sets, no further rewrite-attempt
//   on this term is made.
//
// Global Variables: -
//
// Side Effects    : Changes terms, updates nf_date.
//
/----------------------------------------------------------------------*/

static bool term_li_normalform(RWDesc_p desc, Term_p term)
{
   bool    modified = true, res = false;

   DEBUGMARK(RW_INTERFACE_WATCH2, "TermComputeLINormalform()...\n");   

   /* if(RewriteTermDepth > 15)
      printf("Depth: %ld\n", RewriteTermDepth);*/

   if(desc->level == NoRewrite)
   {
      return false;
   }
   if(SysDateCompare(term->nf_date[desc->level-1], desc->demod_date)!=DateEarlier)
   {
      DEBUGMARK(RW_INTERFACE_WATCH2, "...TermComputeLINormalform() (Used dates)\n");  
      assert(term->weight == TermWeight(term, DEFAULT_VWEIGHT,
					DEFAULT_FWEIGHT));
      return false;
   }      
   if(TermIsVar(term))
   {
      assert(term->weight == DEFAULT_VWEIGHT);
      return false;
   }
   while(modified)
   {
      int  i;
            
      TermGetRef(&term, term);
      /* RewriteTermDepth++;*/
      for(i=0; i<term->arity; i++)
      {
	 modified = term_li_normalform(desc, term->args[i]);
	 res = res || modified;
      }
      /* RewriteTermDepth--; */
      if(!TermIsVar(term)) /* Remember, term itself may have been
			      replaced */
      {
	 term->weight = DEFAULT_FWEIGHT;
	 for(i=0; i<term->arity; i++)
	 {
	    term->weight += term->args[i]->weight;
	 }
      }	
      assert(term->weight == TermWeight(term, DEFAULT_VWEIGHT, DEFAULT_FWEIGHT));
      modified = rewrite_with_clause_setlist(desc->ocb, desc->bank, term,
					     desc->demods, desc->level);
      res = res || modified;
      TermReleaseRef(&term);      
      /* Term may occasionally be without reference here and only
	 lives because of the reference we took above -> we try to
	 delete it to get rid of it. If we do get rid of it, we are,
	 of course, done with this term and it's subterms. */
      if(TBDelete(desc->bank, term))
      {
	 return res;
      }
   }
   term->nf_date[RewriteAdr(RuleRewrite)] = desc->demod_date;
   if(desc->level == FullRewrite)
   {
      term->nf_date[RewriteAdr(FullRewrite)] = desc->demod_date;
   }
   TermCellDelProp(term, TPIsRewritable);
   
   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermComputeLINormalform()\n");   
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: eqn_li_normalform()
//
//   Compute the normal form of maximal, minimal or both terms in an
//   equation. Return true if a term was rewritten.
//
// Global Variables: -
//
// Side Effects    : Changes terms and term bank
//
/----------------------------------------------------------------------*/

bool eqn_li_normalform(RWDesc_p desc,Eqn_p eqn)
{   
   bool res1 = false, res2 = false;

   res1 = term_li_normalform(desc, eqn->lterm);
   res2 = term_li_normalform(desc, eqn->rterm);
			       ;
   return res1 || res2;
}

/*-----------------------------------------------------------------------
//
// Function: rw_desc_cell_alloc()
//
//   Allocate an initialized RWDescCell.
//
// Global Variables: -
//
// Side Effects    : Memory allocation
//
/----------------------------------------------------------------------*/

static __inline__ RWDesc_p rw_desc_cell_alloc(OCB_p ocb, TB_p bank,
					      ClauseSet_p
					      *demodulators,
					      RewriteLevel level) 
{
   RWDesc_p desc = RWDescCellAlloc();
   
   desc->ocb        = ocb;
   desc->bank       = bank;
   desc->demods     = demodulators;
   desc->demod_date = ClauseSetListGetMaxDate(demodulators, level);
   desc->level      = level;
   
   return desc;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TermComputeLINormalform()
//
//   Compute a leftmost-innermost normal form of term. This is more
//   tricky than expected, as supertems on the stack may change. I'll
//   start with a recursive version and see how it works. This uses
//   dates to minimize rewrite-attempts: If the normal form of the
//   term is younger than the clause sets, no further rewrite-attempt
//   on this term is made.
//
// Global Variables: -
//
// Side Effects    : Changes terms, updates nf_date.
//
/----------------------------------------------------------------------*/

bool TermComputeLINormalform(OCB_p ocb, TB_p bank, Term_p term,
			     ClauseSet_p *demodulators, RewriteLevel
			     level)
{
   bool res;
   RWDesc_p desc = rw_desc_cell_alloc(ocb, bank, demodulators, level);

   res = term_li_normalform(desc, term);
   RWDescCellFree(desc);
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnComputeLINormalform()
//
//   Compute the normal form of maximal, minimal or both terms in an
//   equation. Return true if a term was rewritten.
//
// Global Variables: -
//
// Side Effects    : Changes terms and term bank
//
/----------------------------------------------------------------------*/

bool EqnComputeLINormalform(OCB_p ocb, TB_p bank, Eqn_p eqn,
			    ClauseSet_p* demodulators,
			    RewriteLevel level)
{   
   bool res;
   RWDesc_p desc = rw_desc_cell_alloc(ocb, bank, demodulators, level);

   DEBUGMARK(RW_INTERFACE_WATCH1, "EqnComputeLINormalform()...\n");
   res = eqn_li_normalform(desc, eqn);
   RWDescCellFree(desc);
   DEBUGMARK(RW_INTERFACE_WATCH1, "EqnComputeLINormalform()...\n");
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseComputeLINormalform()
//
//   Compute the normal form of terms in a clause. Return true if a
//   term was rewritten. 
//
// Global Variables: -
//
// Side Effects    : Changes terms and term bank
//
/----------------------------------------------------------------------*/

bool ClauseComputeLINormalform(OCB_p ocb, TB_p bank, Clause_p clause,
			       ClauseSet_p *demodulators,
			       RewriteLevel level)
{
   Eqn_p handle;
   bool tmp, res=false;
   RWDesc_p desc = rw_desc_cell_alloc(ocb, bank, demodulators, level);

   /* printf("# ClauseComputeLINormalform(%ld)...\n",clause->ident); */
   DEBUGMARK(RW_INTERFACE_WATCH1, "ClauseComputeLINormalform()...\n");
   DEBUG(RW_INTERFACE_WATCH1, printf("# Norm: ");
	 ClausePrint(stdout, clause, true);printf("\n"););

   for(handle = clause->literals; handle; handle=handle->next)
   {
      tmp = eqn_li_normalform(desc, handle);
      res = res || tmp;
   }
   RWDescCellFree(desc);
   DEBUG(RW_INTERFACE_WATCH1, printf("# Res :");
	 ClausePrint(stdout, clause, true);printf("\n"););
   DEBUGMARK(RW_INTERFACE_WATCH1, "...ClauseComputeLINormalform()\n");   
   /* printf("# ...ClauseComputeLINormalform(%ld)\n",clause->ident); */
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetComputeLINormalform()
//
//   Compute a normal form for terms in all clauses in set with
//   respect to clauses in demodulators up to level. Returns true if a
//   term was rewritten (or res was true initially - res tells the
//   function if any potential terms have been rewritten previously).
//
// Global Variables: -
//
// Side Effects    : Changes clauses
//
/----------------------------------------------------------------------*/

bool ClauseSetComputeLINormalform(OCB_p ocb, TB_p bank, ClauseSet_p
				  set, ClauseSet_p *demodulators,
				  RewriteLevel level, bool res)
{
   Clause_p handle;
   bool     tmp;

   DEBUGMARK(RW_INTERFACE_WATCH1,
	     "ClauseSetComputeLINormalform()...\n");

   for(handle=set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      tmp = ClauseComputeLINormalform(ocb, bank, handle,
				      demodulators, level);
      res = res || tmp;
      if(res&&(ClauseLiteralNumber(handle)>1))
	 /* A term was changed, hence we must assume all
	    clauses have changed! */
      {
	 handle->weight = ClauseStandardWeight(handle);
      }
      assert((ClauseLiteralNumber(handle)<=1) || 
	      (handle->weight == ClauseStandardWeight(handle)));
   }
   DEBUGMARK(RW_INTERFACE_WATCH1, "...ClauseSetComputeLINormalform()\n");
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: FindClausesWithRewritableMaxSides()
//
//   Add pointers to all clauses in set for which (potentially)
//   maximal sides can be rewritten with new_demod to the results
//   tree. This uses two optimizations: 
//
//   1) If a terms nf_date is equal to nf_date, the term (and all
//      subterms) are considered to be irreducible with respect to
//      new_demod (hence nf_date should be bigger than any nf_date of
//      terms in clauses). If a term and all its subterms are
//      irreducible by demod, it's date will be set to nf_date.
//   2) If a terms nf_date is equal to SysDateCreationTime(), the term
//      is considered to be reducible with respect to new_demod. If a
//      term is rewritable, it's nf_date will be set to
//      SysDateCreationTime(). 
//
//   Returns true if any rewritable non-maximal sides are found.
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms
//
/----------------------------------------------------------------------*/

bool FindClausesWithRewritableMaxSides(OCB_p ocb, ClauseSet_p set,
				       PTree_p* results, Clause_p
				       new_demod, SysDate nf_date)
{   
   return find_clauses_with_rw_max_sides(ocb, set, results, new_demod,
					 nf_date);
   /* Later: Use the index if it exists */
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


