/*-----------------------------------------------------------------------

File  : cco_simplification.c

Author: Stephan Schulz

Contents
 
  Control of simplified clauses

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Jun  8 14:49:49 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cco_simplification.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


HCB_p DemodOrderCriterion = NULL;


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: DemodInsert()
//
//   Insert a demodulator into the set...
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void DemodInsert(ClauseSet_p set, Clause_p new)
{
   assert(!new->evaluations);

   if(!set->demod_index)
   {
      ClauseSetInsert(set, new);
   }
   else
   {
      ClauseSetIndexedInsert(set, new);
   }
}


/*-----------------------------------------------------------------------
//
// Function: DemodExtract()
//
//  Extract a demodulator from the set...
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Clause_p DemodExtract(Clause_p handle)
{
   assert(handle->set);

   if(!(handle->set->demod_index))
   {
      assert(!handle->evaluations);
      ClauseSetExtractEntry(handle);
   }
   else
   {
      ClauseSetIndexedExtractEntry(handle);
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseMoveSimplified()
//
//   Remove a processed simplified clause from its set and add it to a
//   temporary set.
//
// Global Variables: -
//
// Side Effects    : Kills children, modifes clause counter
//
/----------------------------------------------------------------------*/

void ClauseMoveSimplified(Clause_p clause, ClauseSet_p tmp_set, long
			  *idents) 
{
   ClauseKillChildren(clause);
   DemodExtract(clause);
   ClauseDeleteTermProperties(clause, TPRestricted);
   ClauseDefaultDocumentModification(clause, CM_TERM_RENAME, NULL);
   ClauseRename(GlobalOut, OutputLevel, clause, ++(*idents));
   ClauseSetInsert(tmp_set, clause);
}

/*-----------------------------------------------------------------------
//
// Function: RemoveClausesWithRewritableMaxSides()
//
//   Remove all clauses for which potentially maximal terms can be
//   rewritten with new_demod and store them into into. Add number
//   of clauses deleted to *count. Return true if a rewritable
//   non-maximal term was encountered.
//
// Global Variables: -
//
// Side Effects    : As specified...
//
/----------------------------------------------------------------------*/

bool RemoveClausesWithRewritableMaxSides(OCB_p ocb, ClauseSet_p from,
					 ClauseSet_p into, Clause_p
					 new_demod, SysDate nf_date,
					 long *idents)
{
   PTree_p  store = NULL;
   PStack_p stack = PStackAlloc();
   Clause_p handle;
   PTree_p  cell;
   bool     res;

   res = FindClausesWithRewritableMaxSides(ocb, from,
					   &store,
					   new_demod,
					   nf_date);
   PStackPushP(stack, store);
   while(!PStackEmpty(stack))
   {
      cell = PStackPopP(stack);
      if(cell)
      {
	 PStackPushP(stack, cell->lson);
	 PStackPushP(stack, cell->rson);
	 handle = cell->key;	 

	 ClauseMoveSimplified(handle, into, idents);
      }
   }
   PStackFree(stack);
   PTreeFree(store);

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetUnitSimplify()
//
//   Try to simplify all clauses in set by performing matching unit
//   resolution with simplifier. Move affected clauses from set into
//   tmp_set. Return number of clauses moved.
//
// Global Variables: -
//
// Side Effects    : Changes clauses and set.
//
/----------------------------------------------------------------------*/

long ClauseSetUnitSimplify(ClauseSet_p set, Clause_p simplifier,
			   ClauseSet_p tmp_set, long *idents)
{
   Clause_p handle, move;
   long res = 0,tmp;
   
   handle = set->anchor->succ; 
   while(handle!=set->anchor)
   {
      tmp = ClauseUnitSimplifyTest(handle, simplifier);
      if(tmp)
      {
	 move = handle;
	 handle = handle->succ;	
	 ClauseMoveSimplified(move, tmp_set, idents);
	 res++;
      }
      else
      {
	 handle = handle->succ;
      }
   }
   return res;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


