/*-----------------------------------------------------------------------

File  : che_heu_auto_mode.c

Author: Stephan Schulz

Contents
 
  Functions implementing a simple auto-mode for E.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Tue Dec  8 17:07:03 MET 1998    
    New

-----------------------------------------------------------------------*/

#include "che_heu_auto_mode.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: create_completion_heuristic()
//
//   Return a heuristic for dealing with standard completion
//   problems. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

HCB_p create_completion_heuristic(HCBARGUMENTS)
{
   HCB_p         hcb     = HCBAlloc();
            
   OUTPRINT(2, "# Optimized rewriting/single goal heuristic selected\n");
   
   HCBAddWFCB(hcb, WFCBAdminFindWFCB(control->wfcbs, "rweight21_gg") ,
	      1);
   
   return hcb;
}

/*-----------------------------------------------------------------------
//
// Function: create_rweight_heuristic()
//
//   Return a heuristic for dealing with equality
//   problems with variables in goals.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

HCB_p create_rweight_heuristic(HCBARGUMENTS)
{
   HCB_p         hcb     = HCBAlloc();

   HCBAddWFCB(hcb, WFCBAdminFindWFCB(control->wfcbs, "rweight21_f") ,
	      1);
 
   return hcb;
}


/*-----------------------------------------------------------------------
//
// Function: create_ex_unit_heuristic()
//
//   Return a heuristic for dealing with unit-problems with
//   existentially quantified goals.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

HCB_p  create_ex_unit_heuristic(HCBARGUMENTS)
{
   HCB_p         hcb     = HCBAlloc();

   OUTPRINT(2, "# Rewriting/Paramod heuristic selected\n");
   
   HCBAddWFCB(hcb, WFCBAdminFindWFCB(control->wfcbs, "auto_1") ,
	      2);
   HCBAddWFCB(hcb, WFCBAdminFindWFCB(control->wfcbs, "weight11_g") ,
	      1);

   return hcb;   
}


/*-----------------------------------------------------------------------
//
// Function: create_default_heuristic()
//
//   Return a simple general-purpose-heuristic.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

HCB_p create_default_heuristic(HCBARGUMENTS)
{
   HCB_p         hcb     = HCBAlloc();
              
   OUTPRINT(2, "# Default heuristic selected\n");
  
   HCBAddWFCB(hcb, WFCBAdminFindWFCB(control->wfcbs, "weight21_g") ,
	      2);
   HCBAddWFCB(hcb, WFCBAdminFindWFCB(control->wfcbs, "weight21_f") ,
	      6);
   HCBAddWFCB(hcb, WFCBAdminFindWFCB(control->wfcbs, "fifo_f") ,
	      1);

   return hcb;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: HCBAutoModeCreate()
//
//   Analyse the proof problem and return an hopefully suitable
//   heuristic. 
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

HCB_p HCBAutoModeCreate(HCBARGUMENTS)
{
   HCB_p res;

   OUTPRINT(1, "\n# Auto-Mode is analysing problem\n");
   if(ClauseSetIsUnitSet(state->axioms))
   {
      OUTPRINT(2, "# Problem is a unit-problem\n");
      if(ClauseSetGoalsAreGround(state->axioms))
      {	 
	 res = create_completion_heuristic(state, control, parms);
      }
      else
      {
	 res = create_ex_unit_heuristic(state, control, parms);
      }
   }
   else if(ClauseSetIsPureEquationalSet(state->axioms))
   {
      OUTPRINT(2, "# Problem is a PEQ problem\n"); 
      res = create_ex_unit_heuristic(state, control, parms);
   }
   else if(ClauseSetIsHornSet(state->axioms))
   {
      OUTPRINT(2, "# Problem is a horn-problem\n");
      
      if(ClauseSetIsEquational(state->axioms))
      {
	 OUTPRINT(2, "# Problem has equality\n");
	 res = create_rweight_heuristic(state, control, parms);
      }
      else
      {
	 OUTPRINT(2, "# Problem has no equality\n");	 
	 res = create_default_heuristic(state, control, parms);	    
      }
      if(ClauseSetNonGoundAxiomPart(state->axioms)>=MANY_DEFAULT)
      {
	 OUTPRINT(2, "# Activating Dershowitz-Hack\n");
	 control->paramod_with_units_only = true;
      }
      else
      {
	 OUTPRINT(2, "# Too few interesting unit-clauses for special treatment \n");
      }
   }
   else
   {
      if(ClauseSetIsEquational(state->axioms))
      {
	 OUTPRINT(2, "# Problem has equality\n");
	 res = create_rweight_heuristic(state, control, parms);
      }
      else
      {
	 OUTPRINT(2, "# Problem has no equality\n");	 
	 res = create_default_heuristic(state, control, parms);	    
      }
   }
   OUTPRINT(1, "# Auto-Mode complete\n");
   
   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


