/*-----------------------------------------------------------------------

File  : cto_ocb.h

Author: Stephan Schulz

Contents
 
  Global definitions for orderings: Comparison results, precedences,
  order control blocks.

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Wed Apr 29 02:51:28 MET DST 1998
    New

-----------------------------------------------------------------------*/

#ifndef CTO_OCB

#define CTO_OCB

#include <cte_termbanks.h>
#include <cto_cmpcache.h>


/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/



typedef enum 
{
   NoOrdering,
   AUTO,
   OPTIMIZE_AX,
   KBO,
   LPO,
   RPO
}TermOrdering;

typedef enum
{
   Multiset,
   LeftRight,
   RightLeft
}CompareStatus;



typedef struct ocb_cell
{
   TermOrdering   type;
   Sig_p          sig;         /* Slightly hacked...this is only an
				  unsupervised reference (but will
				  stay) ! */
   int            sig_size;
   long          *weights;     /* Array of weights */
   long          var_weight;   /* Variable Weight */
   CompareStatus *status;      /* Array of status  */
   CompareResult *precedence;  /* Interpreted as two-dimensional
				   array */
   bool           no_lit_cmp;  /* If true, all literals are
				  uncomparable (useful for SOS
				  strategy) */
   PStack_p       statestack;  /* Contains backtrack information */
   CmpCache_p     cmp_cache;   /* Cache for LPO, RPO and the like... */
}OCBCell, *OCB_p;

#define OCB_FUN_DEFAULT_WEIGHT 1
#define OCB_DEFAULT_STATUS LeftRight

/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

#define OCBCellAlloc() (OCBCell*)SizeMalloc(sizeof(OCBCell))
#define OCBCellFree(junk)        SizeFree(junk, sizeof(OCBCell))

/* Symbolic representation of ordering relations */

extern char*  TONames[];

OCB_p         OCBAlloc(TermOrdering type, Sig_p sig);
void          OCBFree(OCB_p junk);

void          OCBDebugPrint(FILE* out, OCB_p ocb);

PStackPointer OCBPrecedenceAddTuple(OCB_p ocb, FunCode f1, FunCode f2,
				    CompareResult relation);

bool          OCBPrecedenceBacktrack(OCB_p ocb, PStackPointer state);
#define OCBPrecedenceGetState(ocb) \
              PStackGetSP((ocb)->statestack)

/* Getting the addresses of OCB entries for modification */


#define OCBFunWeightPos(ocb, f) \
              (assert((f)>0), assert((f)<=(ocb)->sig_size), &((ocb)->weights[(f)-1]))
#define OCBFunComparePos(ocb, f1, f2) \
              (assert((f1)>0), assert((f2)>0), assert((f1)<=(ocb)->sig_size),\
	       assert((f2)<=(ocb)->sig_size),\
              (&((ocb)->precedence[((f2)-1)*(ocb)->sig_size+((f1)-1)])))
#define OCBFunStatusPos(ocb, f) \
              (assert((f)>0), assert((f)<=(ocb)->sig_size), &((ocb)->status[(f)-1]))

/* Functions for Querying the OCB */

static __inline__ long          OCBFunWeight(OCB_p ocb, FunCode f);
static __inline__ CompareResult OCBFunCompare(OCB_p ocb, FunCode f1, FunCode f2);
static __inline__ CompareStatus OCBFunStatus(OCB_p ocb, FunCode f);


FunCode       OCBTermMaxFunCode(OCB_p ocb, Term_p term);


/*---------------------------------------------------------------------*/
/*                        Inline Functions                             */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: OCBFunWeight()
//
//   Return the weight of f in ocb. For symbols entered in the OCB
//   after creation return OCB_FUN_DEFAULT_WEIGHT.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static __inline__ long OCBFunWeight(OCB_p ocb, FunCode f)
{
   if(f<=ocb->sig_size)
   {
      return *(OCBFunWeightPos(ocb, (f)));
   }
   return OCB_FUN_DEFAULT_WEIGHT;
}


/*-----------------------------------------------------------------------
//
// Function: OCBFunCompare()
//
//   Return comparison result of two symbols in precedence. Symbols
//   not covered by the ocb are smaller than all other versions, and
//   older symbols are smaller than new ones.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static __inline__ CompareResult OCBFunCompare(OCB_p ocb, FunCode f1,
					      FunCode f2)
{
   if(f1<=ocb->sig_size)
   {
      if(f2<=ocb->sig_size)
      {
	 return *(OCBFunComparePos(ocb, f1, f2));
      }
      return to_greater;
   }
   if(f2<=ocb->sig_size)
   {
      return to_lesser;
   }
   assert((f1>ocb->sig_size) && (f1>ocb->sig_size));
   return Q_TO_PART(f2-f1);
}


/*-----------------------------------------------------------------------
//
// Function: OCBFunStatus()
//
//   Return the status of f in ocb. For symbols entered in the OCB
//   after creation return OCB_DEFAULT_STATUS.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static __inline__ CompareStatus OCBFunStatus(OCB_p ocb, FunCode f)
{
   if(f<=ocb->sig_size)
   {
      return *(OCBFunStatusPos(ocb, (f)));
   }
   return OCB_DEFAULT_STATUS;
}

#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





