/*-----------------------------------------------------------------------

File  : cte_termbanks.h

Author: Stephan Schulz

Contents
 
  Definitions for term banks - i.e. shared representations of terms as
  defined in cte_terms.h. Uses the same struct, but adds
  administrative stuff and functionality for sharing.

  There are two sets of funktions for the manangment of term trees:
  Funktions operating only on the top cell, and functions descending
  the term structure. Top level functions implement a conventional AVL
  tree with key f_code.masked_properties.entry_nos_of_args and are
  implemented in cte_termtrees.[ch]

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Sep 22 00:15:39 MET DST 1997
    New
<2> Wed Feb 25 18:16:34 MET 1998
    Adapted for use of new term modules with shared variables


-----------------------------------------------------------------------*/

#ifndef CTE_TERMBANKS

#define CTE_TERMBANKS

#include <clb_numtrees.h>
#include <cio_basicparser.h>
#include <cte_termcellstore.h>

/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/

typedef struct tbcell
{
   TermProperties prop_mask;  /* Which properties are used in this
				 term bank? */ 
   unsigned long in_count;    /* How many terms have been inserted? */
   PDArray_p     ext_index;   /* Associate _external_ abbreviations (=
				 entry_no's with term nodes, necessary
				 for parsing of term bank terms. For
				 critical cases (full protocolls) this
  			         is bound to be densly poulated -> we
				 use an array. Please note that term
				 replacing does not invalidate entries
				 in ext_index
				 (it would be pretty expensive in
				 terms of time and memory), so higher
				 layers have to take care of this if
				 they want to both access terms via
				 references and do replacing! */
   Sig_p         sig;         /* Store sig info */
   VarBank_p     vars;        /* Information about (shared) variables */
   Term_p        true_term;   /* Pointer to the special term with the
				 $true constant. */
   unsigned long rewrite_steps; /* How many calls to TBTermReplace? */
   TermCellStoreCell term_store;  /* Here are the terms */
}TBCell, *TB_p;
   


/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

extern bool TBPrintTermsFlat;
extern bool TBPrintInternalInfo;
extern bool TBSupportReplace;
extern bool TBPrintDetails;

#define TBCellAlloc() (TBCell*)SizeMalloc(sizeof(TBCell))
#define TBCellFree(junk)         SizeFree(junk, sizeof(TBCell))

TB_p    TBAlloc(TermProperties prop_mask, Sig_p sig);
void    TBFree(TB_p junk);

long    TBTermNodes(TB_p bank);
#define TBStorage(bank) (3*((bank)->term_store.entries))

#define TBCellIdent(term) (TermIsVar(term)?(term)->f_code:term->entry_no)

Term_p  DefaultSharedTermCellAlloc(void);
void    TermGetRef(Term_p *ref, Term_p term);
void    TermReleaseRef(Term_p *ref);
void    TermIntroduceSuperterm(Term_p term, Term_p superterm);
bool    TermRemoveSuperterm(Term_p term, Term_p superterm);
#define TermIsTrueTerm(term) ((term)->f_code==SIG_TRUE_CODE)

static __inline__ bool TBTermEqual(Term_p t1, Term_p t2);
bool    TBTermStructEqual(Term_p t1, Term_p t2);
#define TBTermIsSubterm(super, term)\
        TermIsSubterm((super),(term),DEREF_NEVER,TBTermEqual)
Term_p  TBInsert(TB_p bank, Term_p term, DerefType deref);
Term_p  TBInsertNoProps(TB_p bank, Term_p term, DerefType deref);
Term_p  TBInsertInstantiated(TB_p bank, Term_p term);
Term_p  TBTermtopInsert(TB_p bank, Term_p t);
Term_p  TBFind(TB_p bank, Term_p term);
bool    TBDelete(TB_p bank, Term_p term);
void    TBPrintBankInOrder(FILE* out, TB_p bank);
void    TBPrintTermCompact(FILE* out, TB_p bank, Term_p term);
#define TBPrintTermFull(out, bank, term) \
        TermPrint((out), (term), (bank)->sig, DEREF_NEVER)
void    TBPrintTerm(FILE* out, TB_p bank, Term_p term, bool fullterms);
void    TBPrintBankTerms(FILE* out, TB_p bank);
Term_p  TBTermParse(Scanner_p in, TB_p bank);
int     TBTermParseArgList(Scanner_p in, Term_p** arg_anchor,
			  TB_p bank);
long    TBCountSupertermReferences(Term_p term);
long    TBCountTotalReferences(Term_p term);
bool    TBVerifyReferences(FILE* out, Term_p term);
Term_p  TBVerifyBankReferences(FILE* out, TB_p bank);

double  TBAverageSupertermNumber(TB_p bank, double *deviation);
double  TBAverageExtRefNumber(TB_p bank, double *deviation);

void    TBRefSetProp(TB_p bank, TermRef ref, TermProperties prop);
void    TBRefDelProp(TB_p bank, TermRef ref, TermProperties prop);

long    TBTermDelPropCount(Term_p term, TermProperties prop);

long    TBSuperTermsDebugPrint(FILE* out, PTree_p root);

/*-----------------------------------------------------------------------
//
// Function:  TBTermEqual()
//
//   Test wether two shared terms in the same termbank are the
//   same. This is a simple pointer comparison, this function only
//   exists so that it can be passed as a function pointer.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static __inline__ bool TBTermEqual(Term_p t1, Term_p t2)
{
   return (t1==t2);
}


#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





